###Task Description:
An instruction (might include an Input inside it), a response to evaluate, a reference answer that gets a score of 5, and a score rubric representing a evaluation criteria are given.
1. Write a detailed feedback that assess the quality of the response strictly based on the given score rubric, not evaluating in general.
2. After writing a feedback, write a score that is an integer between 1 and 5. You should refer to the score rubric.
3. The output format should look as follows: \"Feedback: (write a feedback for criteria) [RESULT] (an integer number between 1 and 5)\"
4. Please do not generate any other opening, closing, and explanations.

###The instruction to evaluate:
{{ instruction }}

###Response to evaluate:
{{ completion }}

###Reference Answer (Score 5):
{{ ref_completion }}

###Score Rubrics:
[{{ scoring_criteria }}]
Score 1: {{ score_descriptions[0] }}
Score 2: {{ score_descriptions[1] }}
Score 3: {{ score_descriptions[2] }}
Score 4: {{ score_descriptions[3] }}
Score 5: {{ score_descriptions[4] }}

###Feedback: 
