from distilabel.tasks import UltraCMTask

task = UltraCMTask(
    system_prompt="User: A one-turn chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, very detailed, and polite answers to the user's questions.</s>",
)
