{%- if evol_method == "breadth" -%}
I want you to act as a Prompt Creator.
Your goal is to draw inspiration from the #Given Prompt# to create a brand new prompt.
This new prompt should belong to the same domain as the #Given Prompt# but must be even more rare.
The LENGTH and complexity of the #Created Prompt# should be similar to that of the #Given Prompt#.
The #Created Prompt# must be reasonable and must be understood and responded by humans.
'#Given Prompt#', '#Created Prompt#', 'given prompt' and 'created prompt' are not allowed to appear in #Created Prompt#
#Given Prompt#:
{{ instruction }}
#Created Prompt#:
{% else %}{%- if evol_method == "constraints" -%}
{% set depth_rewriter = "Please add one more constraints/requirements into #The Given Prompt#" %}
{%- elif evol_method == "deepen" -%}
{% set depth_rewriter = "If #The Given Prompt# contains inquiries about certain issues, the depth and breadth of the inquiry can be increased." %}
{%- elif evol_method == "concretizing" -%}
{% set depth_rewriter = "Please replace general concepts with more specific concepts." %}
{%- elif evol_method == "reasoning" -%}
{% set depth_rewriter = "If #The Given Prompt# can be solved with just a few simple thinking processes, you can rewrite it to explicitly request multiple-step reasoning." %}
{%- endif -%}
I want you to act as a Prompt Rewriter.
Your objective is to rewrite a given prompt into a more complex version to make those famous AI systems (e.g., chatgpt and GPT4) a bit harder to handle.
But the rewritten prompt must be reasonable and must be understood and responded by humans.
Your rewriting cannot omit the non-text parts such as the table and code in #The Given Prompt#:. Also, please do not omit the input in #The Given Prompt#.
You SHOULD complicate the given prompt using the following method:
{{ depth_rewriter }}
You should try your best not to make the #Rewritten Prompt# become verbose, #Rewritten Prompt# can only add 10 to 20 words into #The Given Prompt#.
'#The Given Prompt#', '#Rewritten Prompt#', 'given prompt' and 'rewritten prompt' are not allowed to appear in #Rewritten Prompt#
#The Given Prompt#:
{{ instruction }}
#Rewritten Prompt#:
{% endif %}