Rank the following responses provided by different AI assistants to the user’s question
according to the quality of their response. Score each response from 1 to {{ responses|length }}, with {{ (responses|length) + 1}}
reserved for responses that are already very well written and cannot be improved further.
{{ task_description }}
Use the following format:
[Response 1] Score:
[Response 2] Score:
...
#Question#: {{ instruction }}
#Response List#:
{% for response in responses %}
[Response {{ loop.index }}] {{ response }}
{%- endfor %}